import os
import pickle
import typing as tp

from satella.coding import silence_excs, Monitor

from smok.extras.event_database import InMemoryEventDatabase
from smok.extras.event_database.in_memory import InMemoryEventSynchronization


class PicklingEventSynchronization(InMemoryEventSynchronization):

    def acknowledge(self, *uuids: str) -> None:
        super().acknowledge(*uuids)
        self.event_db.sync_data()


class PicklingEventDatabase(InMemoryEventDatabase):
    """
    A simple database that pickles events on disk each
    :meth:`~smokclient.extras.BaseEventSynchronization.checkpoint`

    :param cache_path: path to cache with predicate's data
    :param data_path: path to cache with event data
    """

    def __init__(self, cache_path: str, data_path: str):
        super().__init__(cache_path)
        self.data_path = data_path

        if os.path.exists(self.data_path):
            with open(self.data_path, 'rb') as f_in, silence_excs(pickle.PickleError):
                self.events, self.events_to_sync = pickle.load(f_in)

    @Monitor.synchronized
    def get_events_to_sync(self) -> tp.Optional[PicklingEventSynchronization]:
        if not self.events_to_sync:
            return None
        else:
            return PicklingEventSynchronization(self, self.events_to_sync)

    def checkpoint(self):
        self.sync_data()

    @Monitor.synchronized
    def sync_data(self):
        with open(self.data_path, 'wb') as f_out:
            pickle.dump((self.events, self.events_to_sync), f_out)



import base64
import gzip
from logging import LogRecord, Handler

import ujson
from satella.coding import Monitor
from satella.coding.concurrent import SequentialIssuer
from satella.instrumentation import Traceback, frame_from_traceback
from satella.time import time_us

__all__ = ['SMOKLogHandler']


class SMOKLogHandler(Handler, Monitor):
    """
    A logging handler for SMOK

    :param device: SMOKDevice that this logger is attached to
    :param service_name: name of the service
    """
    def __init__(self, device: 'SMOKDevice', service_name: str):
        super().__init__()
        Monitor.__init__(self)

        self.service_name = service_name
        self.device = device
        # So timestamps inserted will be monotonically increasing
        self.timestamper = SequentialIssuer(time_us())

    def record_to_json(self, record: LogRecord):
        ts = self.timestamper.no_less_than(time_us())

        # noinspection PyBroadException
        try:
            msg = self.format(record)
        except (TypeError, ValueError):
            msg = record.message + ' '+','.join(map(repr, record.args))

        if len(msg) > 1000:
            msg_content = base64.b64encode(gzip.compress(msg.encode('utf8'), compresslevel=8)).decode('utf8')
            msg = {'encoding': 'base64-gzip', 'content': msg_content}

        dct = {
            'service': self.service_name,
            'when': ts,
            'message': msg,
            'level': record.levelno,
        }

        if record.exc_info:
            f = frame_from_traceback(record.exc_info[2])
            tb = Traceback(f)
            tb_json = base64.b64encode(gzip.compress(ujson.dumps(tb.to_json()).encode('utf-8'), 9)).decode('utf-8')
            dct.update(exception_text=tb.pretty_format(),
                       exception_traceback=tb_json)

        return dct

    def emit(self, record: LogRecord) -> None:
        self.device.log_publisher.queue.put(self.record_to_json(record))

    def close(self) -> None:
        pass




import typing as tp

from .base import BaseEventDatabase, BaseEventSynchronization
from ...predicate import Event


class NullEventDatabase(BaseEventDatabase):
    """
    For these use cases where you don't use events at all.
    """

    def on_predicate_deleted(self, predicate_id: str) -> None:
        pass

    def get_cache(self, predicate_id: str) -> tp.Any:
        raise KeyError()

    def set_cache(self, predicate_id: str, cache) -> None:
        pass

    def get_events_to_sync(self) -> tp.Optional[BaseEventSynchronization]:
        return None

    def add_event(self, event: Event) -> None:
        pass

    def close_event(self, event: Event) -> None:
        pass

    def get_open_events(self) -> tp.Iterator[Event]:
        return iter([])

    def checkpoint(self) -> None:
        pass



import time
import typing as tp
import queue
from urllib.parse import urlencode, quote_plus

from satella.coding import for_argument, queue_get
from satella.coding.decorators import retry
from satella.coding.structures import CacheDict

from smok.exceptions import ResponseError
from smok.extras import BaseMetadataDatabase


class PlainDataUpdater:
    def __init__(self, device: 'SMOKDevice'):
        self.entries = queue.Queue()
        self.device = device

    def add_update(self, key: str, value: str, timestamp: tp.Optional[float] = None):
        timestamp = timestamp or time.time()
        self.entries.put((key, value, timestamp))

    @queue_get('entries', 0)
    @retry(3, exc_classes=ResponseError)
    def try_update(self, msg):
        key, value, timestamp = msg

        if not value:
            self.device.api.delete('/v1/device/metadata/plain/%s' % (urlencode(key), ))
        else:
            self.device.api.put('/v1/device/metadata/plain/%s' % (quote_plus(key), ),
                                json={'value': value, 'unless_timestamp_higher': timestamp})


class StorageDict:
    def __init__(self, device: 'SMOKDevice'):
        self.db = device.meta_database  # type: BaseMetadataDatabase

    def __getitem__(self, item: str) -> str:
        return self.db.get_plain(item)

    @for_argument(None, None, str)
    def __setitem__(self, key: str, value: str):
        self.db.set_plain(key, value)

    def __delitem__(self, key):
        self.db.delete_plain(key)


CACHE_MAX_FOR = 60


class PlainMetadata(CacheDict):
    """
    An object representing device's plain metadata.

    Entries will be cached for 60 seconds after downloading them from the server.

    Note that this does not support iteration or getting length, but supports
    __setitem__, __delitem__ and __getitem__.
    """
    __slots__ = ('device', 'updater', 'db', 'cached_for')

    def try_update(self):
        self.updater.try_update()

    def __init__(self, device: 'SMOKDevice'):
        self.updater = PlainDataUpdater(device)
        self.db = device.meta_database
        self.device = device
        self.cached_for = {}

    def __getitem__(self, item):
        try:
            cached_for = self.cached_for[item]
            if time.monotonic() - cached_for > self.device.cache_metadata_for:
                raise KeyError()
            return self.db.get_plain(item)
        except KeyError:
            try:
                resp = self.device.api.get('/v1/device/metadata/plain/%s' % (
                    quote_plus(item)
                ))
                value = resp['value']
                self.db.update_plain(item, value, resp['timestamp'])
                self.cached_for[item] = time.monotonic()
                return value
            except ResponseError as e:
                if e.status_code == 404:
                    raise KeyError('key not found')
                return self.db.get_plain(item)

    def get(self, key: str, default=None):
        """
        Get a value for given metadata, returning default if it KeyErrors
        """
        try:
            return self[key]
        except KeyError:
            return default

    @for_argument(None, None, str)
    def __setitem__(self, key, value):
        self.db.put_plain(key, value)
        self.updater.add_update(key, value, time.time())
        self.cached_for[key] = time.monotonic()

    def __delitem__(self, key):
        self.db.delete_plain(key)
        self.updater.add_update(key, None, time.time())
        try:
            del self.cached_for[key]
        except KeyError:
            pass




import typing as tp


def parse(invocation: str) -> tp.Tuple[str, tp.List[str]]:  # Parse Reparse-Point-type invocation
    """
    Parses reparse point variable code

    Invocation is ex. {abc}+12+{def}, where abc and def are paths.
    Their values would get substituted in place of those {}'s.
    Of course in resulting Python expressions you will have to make
    your dictionaries by dict() constructor, then.

    :param invocation: part of the path without 'r' and without it's type
    :return: tuple. First element is a Python expression. Aforementioned
        example would be parsed to v0+12+v1.
        Second element is a list of pathes of mentioned variables
    :raise ValueError: the path given was invalid
    """
    if not invocation:
        raise ValueError('Path is empty')

    paths = []
    expression = ''

    state = False  # Whether we are scanning a variable instead of a literal

    nesting = 0  # so that stuff like {a{1{2{3}}}} will be parsed
    # as 'a{1{2{3}}}' and not 'a{1{2{3'
    c_var = ''

    for c in invocation:
        if not state:  # scanning literal mode
            if c == '{':
                state = True
                nesting = 1
            else:
                expression += c
        else:  # scanning variable mode
            if c == '}':
                nesting -= 1
                if not nesting:
                    paths.append(c_var)
                    expression += 'v' + str(int(len(paths) - 1))
                    state = False
                    c_var = ''
                    continue
            elif c == '{':
                nesting += 1

            c_var += c

    if nesting:
        raise ValueError('No matching } found')

    return expression, paths


def reparse_to_native_components(path: str) -> tp.List[str]:
    """
    Analyze a reparse pathpoint and break it down to it's native components.
    Ie. if a reparse pathpoint consist of reparse pathpoints, it will be broken down
    as deep as possible.
    """
    if path[0] != 'r':
        return [path]
    paths = parse(path)[1]
    output = []
    for sub_path in paths:
        output.extend(reparse_to_native_components(sub_path))
    return output



import typing as tp
from abc import ABCMeta, abstractmethod

from smok.exceptions import OperationFailedError
from smok.pathpoint import PathpointValueType, PathpointType
from smok.pathpoint.typing import to_type

SensorValueType = tp.Union[str, int, float, list, dict]
SVTOrExcept = tp.Union[SensorValueType, OperationFailedError]


class BasicType(metaclass=ABCMeta):
    """
    A base class for all SMOK sensor types
    """
    __slots__ = ()

    @abstractmethod
    def pathpoint_to_sensor(self, *values: PathpointValueType) -> SensorValueType:
        """
        Convert a value from pathpoint values to a sensor value
        """

    @abstractmethod
    def sensor_to_pathpoint(self, value: SensorValueType,
                            *pathpoint_names: str) -> tp.Tuple[PathpointValueType, ...]:
        """
        Convert a value from sensor values to pathpoint values, each value in tuple
        for a separate pathpoint. If your sensor consists of a single pathpoint, then
        you should return a 1-element tuple.
        """


class NumericType(BasicType):
    """
    A type for values of numeric type

    :ivar precision: number of places after the comma that should be diplayed (int)
    """
    __slots__ = ('precision', 'multiplier', 'offset')

    def __init__(self, precision=2, multiplier=1, offset=0, **kwargs):
        self.precision = precision
        self.multiplier = multiplier
        self.offset = offset

    def pathpoint_to_sensor(self, *values: PathpointValueType):
        value = values[0]
        value = value * self.multiplier + self.offset
        return round(value, self.precision)

    def sensor_to_pathpoint(self, value: SensorValueType,
                            *pathpoint_names: str) -> tp.Tuple[PathpointValueType, ...]:
        value = (value - self.offset) / self.multiplier
        return to_type(value, PathpointType.get_type(pathpoint_names[0])),


class UnicodeType(BasicType):
    """
    A basic type for Unicode-containing sensors
    """
    def sensor_to_pathpoint(self, value: SensorValueType,
                            *pathpoint_names: str) -> tp.Tuple[PathpointValueType, ...]:
        return str(value),

    def pathpoint_to_sensor(self, *values: PathpointValueType) -> SensorValueType:
        return str(values[0])


CACHE_DICT = {}  # type: tp.Dict[str, NumericType]


def get_type(type_name: str) -> NumericType:
    global CACHE_DICT

    if type_name not in CACHE_DICT:
        if type_name == 'std.Number10':
            type_ = NumericType(multiplier=0.1)
        elif type_name == 'std.Unicode':
            type_ = UnicodeType()
        elif type_name == 'frisko.DayOfWeek':
            type_ = NumericType(offset=-1)
        elif type_name == 'std.Number100':
            type_ = NumericType(multiplier=0.01)
        elif '(' in type_name:
            arguments = type_name.split('(', 1)[1]
            arguments = arguments.rsplit(')', 1)[0]
            type_ = eval('NumericType(%s)' % (arguments,), {'NumericType': NumericType})
        else:
            type_ = NumericType()
        CACHE_DICT[type_name] = type_

    return CACHE_DICT[type_name]



import typing as tp
from abc import ABCMeta, abstractmethod

from satella.coding.typing import Number

from smok.pathpoint.typing import ValueOrExcept, PathpointValueType


class BaseDataToSynchronize(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def acknowledge(self) -> None:
        """
        Mark the data as successfully synchronized
        """

    @abstractmethod
    def negative_acknowledge(self) -> None:
        """
        Mark the data as failed to synchronize
        """

    @abstractmethod
    def to_json(self) -> tp.List:
        """
        Return a JSON structure that looks like this
        (specification expressed in OpenAPI 3.0 format)

        ::

            type: array
            items:
                type: object
                properties:
                    path:
                        type: string
                        description: Name of the pathpoint to serve
                    values:
                        type: array
                        items:
                            type: object
                            description: This has either error_code or value
                            properties:
                                timestamp:
                                    type: integer
                                    format: int64
                                    description: Timestamp in milliseconds
                                error_code:
                                    type: string
                                    description: Reason of error code
                                    enum:
                                        - malformed
                                        - timeout
                                        - invalid
                                value:
                                    type:
                                        oneOf:
                                            - integer
                                            - string
                                            - number
                                    description: Value of the pathpoint
                            required:
                                - timestamp
       """


class BasePathpointDatabase(metaclass=ABCMeta):
    """
    An abstract base class for pathpoint archives database.

    Try to make it reasonably thread-safe. It is documented which call is called by which thread.
    """
    __slots__ = ()

    @abstractmethod
    def checkpoint(self) -> None:
        """
        Called by the communicator thread, once every about 20 seconds.
        """

    @abstractmethod
    def on_new_data(self, pathpoint: str, timestamp: Number,
                    value_or_exception: ValueOrExcept) -> None:
        """
        Called whenever there's a new value.

        Called by order executor thread and user-threads if you make use of
        method :meth:`~smokclient.pathpoint.Pathpoint.set_new_value`.

        :param pathpoint: :term:`Native` pathpoint that has been written
        :param timestamp: timestamp of the operation in milliseconds
        :param value_or_exception: a value of the pathpoint or an OperationFailedError instance
        """

    @abstractmethod
    def get_current_value(self, pathpoint: str) -> tp.Tuple[Number, PathpointValueType]:
        """
        Get the current value for given pathpoint.

        Called by user threads.

        :param pathpoint: name of the pathpoint
        :return: a tuple of timestamp, value
        :raises OperationFailedError: read of this pathpoint has failed
        :raises NotReadedError: pathpoint has no last value (or does not exist)
        """

    @abstractmethod
    def get_archive_data(self, pathpoint: str, start: tp.Optional[Number],
                         stop: tp.Optional[Number]) -> tp.Iterator[tp.Tuple[Number, ValueOrExcept]]:
        """
        Return some archival data for this pathpoint

        Called by user threads.

        :param pathpoint: name of the pathpoint
        :param start: start of the period, in milliseconds since Epoch,
            or the earliest timestamp that is registered if None
        :param stop: stop of the period, in milliseconds since Epoch,
            or the latest timestamp that is registered if None
        :return: an iterator of two-tuple (timestamp in milliseconds, value or
            OperationReadError instance)
        """

    @abstractmethod
    def get_data_to_sync(self) -> tp.Optional[BaseDataToSynchronize]:
        """
        At most one instance of BaseDataToSynchronize is guaranteed to exist at given time.

        Called by communicator thread.

        :return: return data to synchronize, or None if nothing is to synchronize
        """



import typing as tp
import queue

from satella.coding import queue_get
from satella.coding.concurrent import TerminableThread
from satella.coding.decorators import retry

from ..exceptions import ResponseError


MAX_SYNC_AT_ONCE = 50


class LogPublisherThread(TerminableThread):
    def __init__(self, device: 'SMOKDevice'):
        super().__init__(name='log publisher')
        self.device = device
        self.queue = queue.Queue()

    def get_all_messages(self, starting_msg: dict) -> tp.List[dict]:
        msgs = [starting_msg]
        while self.queue.qsize() and len(msgs) < MAX_SYNC_AT_ONCE:
            msgs.append(self.queue.get())
        return msgs

    @queue_get('queue', timeout=5)
    def loop(self, msg):
        msgs = self.get_all_messages(msg)
        self.sync(msgs)

    @retry(3, exc_classes=ResponseError)
    def sync(self, lst: tp.List[dict]):
        self.device.api.put('/v1/device/device_logs', json=lst, timeout=20)

    def cleanup(self):
        while self.queue.qsize() > 0:
            msgs = self.get_all_messages(self.queue.get())
            self.sync(msgs)



import logging
import queue
import time
import typing as tp

import ujson
from satella.coding import silence_excs, for_argument, log_exceptions
from satella.coding.concurrent import TerminableThread, Condition
from satella.coding.decorators import retry
from satella.coding.transforms import jsonify
from satella.exceptions import WouldWaitMore
from satella.time import measure

from smok.basics import StorageLevel
from smok.exceptions import ResponseError
from smok.extras.event_database import BaseEventSynchronization
from smok.pathpoint.data_sync_dict import DataSyncDict
from smok.pathpoint.orders import sections_from_list
from smok.pathpoint.pathpoint import Pathpoint
from smok.predicate import DisabledTime
from smok.predicate.undefined import UndefinedStatistic
from smok.sensor import Sensor

logger = logging.getLogger(__name__)

SENSORS_SYNC_INTERVAL = 300


@for_argument(returns=jsonify)
def pathpoints_to_json(pps: tp.Iterable[Pathpoint]) -> list:
    output = []
    for pp in pps:
        output.append({'path': pp.name,
                       'storage_level': pp.storage_level})
    return output


COMMUNICATOR_INTERVAL = 20
PREDICATE_SYNC_INTERVAL = 300
BAOB_SYNC_INTERVAL = 60*60  # an hour


class CommunicatorThread(TerminableThread):
    def __init__(self, device: 'SMOKClient', order_queue: queue.Queue,
                 data_to_sync: DataSyncDict, dont_obtain_orders: bool,
                 dont_do_baobs: bool, dont_do_pathpoints: bool,
                 dont_do_predicates: bool,
                 startup_delay: float):
        super().__init__(name='order getter')
        self.device = device
        self.startup_delay = startup_delay
        self.dont_do_pathpoints = dont_do_pathpoints
        self.dont_do_baobs = dont_do_baobs
        self.dont_obtain_orders = dont_obtain_orders
        self.queue = order_queue
        self.dont_do_predicates = dont_do_predicates
        self.data_to_sync = data_to_sync
        self.last_sensors_synced = 0
        self.last_predicates_synced = 0
        self.data_to_update = Condition()
        self.last_baob_synced = 0

    def tick_predicates(self) -> None:
        for predicate in self.device.predicates.values():
            # noinspection PyProtectedMember
            predicate._call_method('on_tick')

    @retry(3, ResponseError)
    def sync_events(self) -> None:
        evt_to_sync = self.device.evt_database.get_events_to_sync()  # type: BaseEventSynchronization
        if evt_to_sync is None:
            return
        try:
            resp = self.device.api.post('/v1/device/alarms',
                                        json=[evt.to_json() for evt in evt_to_sync.get_events()])
            evt_to_sync.acknowledge(*(data['uuid'] for data in resp))
        except ResponseError as e:
            logger.error('Failure syncing events', exc_info=e)
            evt_to_sync.negative_acknowledge()
            raise

    @retry(3, ResponseError)
    def sync_predicates(self) -> None:
        resp = self.device.api.get('/v1/device/predicates')

        predicates_found = set()
        for predicate_dict in resp:
            predicate_id = predicate_dict['predicate_id']
            predicates_found.add(predicate_id)
            if not predicate_dict['online']:
                if predicate_id in self.device.predicates:
                    self.device.predicates[predicate_id].on_offline()
                    del self.device.predicates[predicate_id]
            else:
                silencing = [DisabledTime.from_json(dct) for dct in
                             predicate_dict.get('silencing', ())]

                if predicate_id not in self.device.predicates:
                    stat_name = predicate_dict['statistic']
                    stat_class = self.device.predicate_classes.get(stat_name, UndefinedStatistic)
                    predicate = stat_class(self.device, predicate_id,
                                           predicate_dict['verbose_name'], silencing,
                                           predicate_dict['configuration'],
                                           predicate_dict['statistic'])
                    self.device.predicates[predicate_id] = predicate
                else:
                    stat = self.device.predicates[predicate_id]
                    config = predicate_dict['configuration']
                    if stat.configuration != config:
                        stat.on_configuration_changed(config)
                    if stat.silencing != silencing:
                        stat.on_silencing_changed(silencing)

            predicates_to_delete = set(self.device.predicates.keys()) - predicates_found
            for predicate_id in predicates_to_delete:
                self.device.predicates[predicate_id].on_offline()
                del self.device.predicates[predicate_id]

        if not self.last_predicates_synced:
            self.device.ready_lock.release()
        self.last_predicates_synced = time.time()

    @retry(3, ResponseError)
    def sync_sensors(self) -> None:
        resp = self.device.api.get('/v1/device/sensors')

        self.device.sensor_database.on_sensors_sync([Sensor.from_json(self.device, data) for data in resp])
        self.last_sensors_synced = time.time()

    @retry(3, ResponseError)
    def fetch_orders(self) -> None:
        resp = self.device.api.post('/v1/device/orders')
        if resp:
            for section in sections_from_list(resp):
                self.queue.put(section)

    @retry(3, ResponseError)
    @log_exceptions(logger, logging.WARNING, 'Failed to sync data: {e}')
    def sync_data(self) -> None:
        sync = self.data_to_sync.get_data_to_sync()
        if sync is None:
            return
        try:
            data = sync.to_json()
            if not data:
                sync.acknowledge()
                return
            self.device.api.post('/v1/device/pathpoints', json=data)
            sync.acknowledge()
        except ResponseError as e:
            if e.status_code // 100 == 5:
                logger.debug(f'Failed to sync data', exc_info=e)
                sync.negative_acknowledge()
                raise
            else:
                logger.warning(str(data))
                logger.warning(f'Got HTTP {e.status_code} while syncing pathpoint data. '
                               f'Assuming is it damaged and marking as synced')
                sync.acknowledge()

    def sync_baob(self):
        self._sync_baob()
        self.device.baobs_loaded = True

    @retry(3, ResponseError)
    def _sync_baob(self) -> None:
        keys = self.device.baob_database.get_all_keys()
        data = []
        for key in keys:
            data.append({'key': key,
                         'version': self.device.baob_database.get_baob_version(key)})
        data = self.device.api.post('/v1/device/baobs', json=data)
        for key_to_download in data['should_download']:
            resp, headers = self.device.api.get(f'/v1/device/baobs/{key_to_download}',
                                                direct_response=True)
            self.device.baob_database.set_baob_value(key_to_download, resp, int(headers['X-SMOK-BAOB-Version']))
            logger.debug('Downloaded BAOB %s', key_to_download)
        for key_to_upload in data['should_upload']:
            self.device.api.put(f'/v1/device/baobs/{key_to_upload}', files={
                'file': self.device.baob_database.get_baob_value(key_to_upload),
                'data': ujson.dumps({'version': self.device.baob_database.get_baob_version(key_to_upload)}).encode('utf8')
            })
            logger.debug('Uploaded BAOB %s', key_to_upload)
        self.last_baob_synced = time.time()

    @retry(3, ResponseError)
    def sync_pathpoints(self) -> None:
        pps = self.device.pathpoints.copy_and_clear_dirty()
        data = pathpoints_to_json(pps.values())
        try:
            resp = self.device.api.put('/v1/device/pathpoints', json=data)
            for pp in resp:
                name = pp['path']
                if name.startswith('r'):  # Don't use reparse pathpoints
                    continue
                with silence_excs(KeyError):
                    pathpoint = self.device.provide_unknown_pathpoint(name, StorageLevel(
                        pp.get(
                            'storage_level',
                            1)))
                    stor_level = StorageLevel(pp.get('storage_level', 1))
                    if stor_level != pathpoint.storage_level:
                        pathpoint.on_new_storage_level(stor_level)
        except ResponseError:
            self.device.pathpoints.update(pps)
            self.device.pathpoints.dirty = True
            raise

    def prepare(self) -> None:
        # Give the app a moment to prepare and define it's pathpoints
        self.safe_sleep(self.startup_delay)

    def loop(self) -> None:
        with measure() as measurement:
            # Synchronize the data
            if not self.dont_do_pathpoints:
                self.sync_data()

                # Synchronize the pathpoints
                if self.device.pathpoints.dirty:
                    self.sync_pathpoints()

                # Synchronize sensors
                if time.time() - self.last_sensors_synced > SENSORS_SYNC_INTERVAL:
                    self.sync_sensors()

            # Synchronize predicates
            if not self.dont_do_predicates:
                if time.time() - self.last_predicates_synced > PREDICATE_SYNC_INTERVAL:
                    self.sync_predicates()

            # Fetch the BAOBs
            if not self.dont_do_baobs:
                if time.time() - self.last_baob_synced > BAOB_SYNC_INTERVAL:
                    self.sync_baob()

            # Fetch the orders
            if not self.dont_obtain_orders:
                self.fetch_orders()

            if not self.dont_do_predicates:
                # Tick the predicates
                self.tick_predicates()

                # Sync the events
                self.sync_events()

            # Checkpoint the DB
            if not self.dont_do_pathpoints:
                self.device.pp_database.checkpoint()
            if not self.dont_do_predicates:
                self.device.evt_database.checkpoint()

            # Wait for variables to refresh, do we need to upload any?
            time_to_wait = COMMUNICATOR_INTERVAL - measurement()
            while time_to_wait > 0.1:       # for float roundings
                try:
                    ttw = min(time_to_wait, 5)
                    self.data_to_update.wait(timeout=ttw)
                    break
                except WouldWaitMore:
                    time_to_wait -= ttw * 1.1



import time
import typing as tp

from satella.coding import silence_excs

from .base import BaseMetadataDatabase


class InMemoryMetadataDatabase(BaseMetadataDatabase):
    """
    Base class for metadata databases
    """

    @silence_excs(KeyError)
    def delete_plain(self, key: str) -> None:
        del self.db_plain[key]

    @silence_excs(KeyError)
    def delete_set(self, key: str) -> None:
        del self.db_set[key]

    def __init__(self):
        self.db_plain = {}
        self.db_set = {}

    def put_plain(self, key: str, value: str, timestamp: tp.Optional[float] = None) -> None:
        self.db_plain[key] = value, timestamp or time.time()

    def get_plain(self, key: str) -> str:
        return self.db_plain[key][0]

    def update_plain(self, key: str, value: str, timestamp: float) -> None:
        if key not in self.db_plain:
            self.db_plain[key] = value, timestamp
        elif self.db_plain[key][1] < timestamp:
            self.db_plain[key] = value, timestamp

    def get_all_plain(self) -> tp.Iterator[tp.Tuple[str, str, float]]:
        for key, row in self.db_plain.items():
            yield (key, *row)




import logging
import enum
import time
import typing as tp
from concurrent.futures import Future

from satella.coding.structures import ReprableMixin

from smok.pathpoint.typing import PathpointValueType

__all__ = ['AdviseLevel', 'Disposition', 'Order', 'ReadOrder', 'WriteOrder',
           'WaitOrder', 'MessageOrder', 'Section', 'sections_from_list']

logger = logging.getLogger(__name__)


class AdviseLevel(enum.IntEnum):
    """
    Advise level specifies how hard should the device try to execute this command.
    """
    ADVISE = 0  #: Best-effort
    FORCE = 1  #: Nearly guarantees correct delivery, up to blocking the pipeline if need be


class Order:
    """Base class for all orders"""
    __slots__ = ()

    def __repr__(self) -> str:
        return str(type(self))

    def __str__(self) -> str:
        return str(type(self))


class SysctlOrder(Order, ReprableMixin):
    """
    A sysctl order. These are completely user-defined.
    """
    _REPR_FIELDS = ('op_type', 'op_args')
    __slots__ = ('op_type', 'op_args')

    def __init__(self, op_type: str, op_args: str):
        self.op_type = op_type
        self.op_args = op_args

    @classmethod
    def from_json(cls, dct: dict) -> 'SysctlOrder':
        return SysctlOrder(dct['op_type'], dct['op_args'])


class MessageOrder(Order, ReprableMixin):
    """
    A message order. Best executed with
    :meth:`smok.client.SMOKDevice._execute_message_order`
    """
    _REPR_FIELDS = ('uuid', 'times_retry')
    __slots__ = ('uuid', 'times_retry')

    def __init__(self, uuid: str):
        self.uuid = uuid
        self.times_retry = 3

    def fail(self) -> bool:
        self.times_retry -= 1
        return bool(self.times_retry)

    @classmethod
    def from_json(cls, dct: dict) -> 'MessageOrder':
        return MessageOrder(dct['uuid'])


class WaitOrder(Order, ReprableMixin):
    """
    Order to wait a given amount of seconds

    :param period: seconds to wait
    """
    _REPR_FIELDS = ('period',)
    __slots__ = ('period',)

    def __init__(self, period: float):
        self.period = period

    @classmethod
    def from_json(cls, dct: dict) -> 'WaitOrder':
        return WaitOrder(dct['time'])


class WriteOrder(Order, ReprableMixin):
    """
    Order to write a target value to target pathpoint

    :param pathpoint: name of pathpoint to write to
    :param value: value to write
    :param advise: advise level
    :param stale_after: optional timestamp in seconds, after which this write
        will be discarded
    """
    _REPR_FIELDS = ('pathpoint', 'value', 'advise', 'stale_after')
    __slots__ = ('pathpoint', 'value', 'advise', 'stale_after', 'repeat_count')

    def __init__(self, pathpoint: str, value: PathpointValueType, advise: AdviseLevel,
                 stale_after: tp.Optional[float] = None):
        self.pathpoint = pathpoint
        self.value = value
        self.advise = advise
        self.repeat_count = 10 if advise == AdviseLevel.FORCE else 1
        self.stale_after = stale_after

    def is_valid(self) -> bool:
        if self.stale_after is None:
            return True
        return self.stale_after > time.time()

    def fail(self) -> bool:
        """
        Fail this order.

        Return whether to requeue it
        """
        self.repeat_count -= 1
        return bool(self.repeat_count)

    @classmethod
    def from_json(cls, dct: dict) -> 'WriteOrder':
        return WriteOrder(dct['path'], dct['value'], AdviseLevel(dct.get('advise', 0)),
                          dct.get('stale_after'))


class ReadOrder(Order, ReprableMixin):
    """
    An order to read a pathpoint

    :param pathpoint: pathpoint to read
    :param advise: advise level
    """
    _REPR_FIELDS = ('pathpoint', 'advise')

    def __init__(self, pathpoint: str, advise: AdviseLevel):
        self.pathpoint = pathpoint
        self.advise = advise
        self.repeat_count = 3 if AdviseLevel.ADVISE else 20

    @classmethod
    def from_json(cls, dct: dict) -> 'ReadOrder':
        return ReadOrder(dct['path'], AdviseLevel(dct.get('advise', 0)))

    def fail(self) -> bool:
        """
        Fail this order.

        Return whether to requeue it
        """
        self.repeat_count -= 1
        return bool(self.repeat_count)


def orders_from_list(lst: tp.List[dict]) -> tp.List[Order]:
    orders = []
    for order in lst:
        order_type = order['type']
        if order_type == 'message':
            o = MessageOrder.from_json(order)
        elif order_type == 'read':
            o = ReadOrder.from_json(order)
        elif order_type == 'sysctl':
            o = SysctlOrder.from_json(order)
        elif order_type == 'wait':
            o = WaitOrder.from_json(order)
        elif order_type == 'write':
            o = WriteOrder.from_json(order)
        else:
            logger.info('Received unknown order type of %s, ignoring', order_type)
            continue

        if o:
            orders.append(o)
    return orders


class Disposition(enum.IntEnum):
    """
    A joinable quality of the section
    """
    JOINABLE = 0        #: this section can be joined to neighbouring JOINABLE sections
    CANNOT_JOIN = 1     #: this section cannot be joined to any sections


class Section(ReprableMixin):
    """
    A collection of orders.

    :param orders: a list of orders
    :param disposition: if Disposition.JOINABLE then this section can be joined with
        other sections. If Disposition.CANNOT_JOIN then all orders from this section
        will be executed before proceeding to next one
    """
    _REPR_FIELDS = ('orders', 'disposition')
    __slots__ = ('orders', 'disposition', 'future')

    def __init__(self, orders: tp.List[Order] = None,
                 disposition: Disposition = Disposition.JOINABLE):
        self.future = Future()
        self.orders = orders or []
        self.disposition = disposition

    def cancel(self) -> bool:
        return self.future.cancel()

    def result(self, timeout: tp.Optional[float] = None):
        self.future.result(timeout)

    def __bool__(self) -> bool:
        return bool(self.orders)

    @classmethod
    def from_json(cls, dct: dict):
        return Section(orders_from_list(dct['orders']), Disposition(dct.get('disposition', 0)))

    def __iadd__(self, other: tp.Union[Order, 'Section']) -> 'Section':
        if isinstance(other, Order):
            self.orders.append(other)
        else:
            other.future = self.future
            self.orders.extend(other.orders)
        return self

    def is_joinable(self) -> bool:
        return self.disposition == Disposition.JOINABLE

    def max_wait(self) -> tp.Optional[float]:
        wait = None
        for order in (or_ for or_ in self.orders if isinstance(or_, WaitOrder)):
            if wait is None:
                wait = order.period
            else:
                if wait < order.period:
                    wait = order.period
        return wait


def sections_from_list(lst: tp.List[dict]) -> tp.List[Section]:
    return [Section.from_json(section) for section in lst]



from .event_database import BaseEventDatabase, BaseEventSynchronization, PicklingEventDatabase, \
    NullEventDatabase
from .pp_database import BasePathpointDatabase, BaseDataToSynchronize, PicklingPathpointDatabase
from .macros_database import BaseMacroDatabase, PicklingMacroDatabase
from .metadata_database import BaseMetadataDatabase, PicklingMetadataDatabase
from .sensors_database import BaseSensorDatabase, PicklingSensorDatabase
from .baob_database import BAOBDigest, BaseBAOBDatabase, PicklingBAOBDatabase
from .arch_database import BaseArchivesDatabase, PicklingArchivesDatabase

__all__ = ['BasePathpointDatabase', 'BaseDataToSynchronize', 'PicklingPathpointDatabase',
           'BaseEventDatabase', 'BaseEventSynchronization', 'PicklingEventDatabase',
           'BaseMacroDatabase', 'PicklingMacroDatabase', 'PicklingBAOBDatabase',
           'BaseMetadataDatabase', 'PicklingMetadataDatabase',
           'BaseSensorDatabase', 'PicklingSensorDatabase',
           'BAOBDigest', 'BaseBAOBDatabase', 'NullEventDatabase',
           'BaseArchivesDatabase', 'PicklingArchivesDatabase']



import time
import typing as tp

from satella.coding import Monitor
from satella.coding.concurrent import Condition
from satella.coding.structures import DirtyDict

from .typing import PathpointValueType
from ..exceptions import OperationFailedError, OperationFailedReason


class DataSyncDict(DirtyDict, Monitor):
    def __init__(self):
        super().__init__()
        Monitor.__init__(self)
        self.updated_condition = Condition()

    def on_readed_successfully(self, pathpoint: str, value: PathpointValueType,
                               timestamp: tp.Optional[float] = None):
        self.append(pathpoint, timestamp or time.time(), value)

    @Monitor.synchronized
    def append(self, pathpoint, timestamp, value):
        tpl = (timestamp, value)
        if pathpoint not in self:
            self[pathpoint] = [tpl]
        else:
            ts_last = self[pathpoint][-1][0]
            if timestamp > ts_last:
                self[pathpoint].append(tpl)
        self.dirty = True
        self.updated_condition.notify()

    def add_from_json(self, dct: tp.List[dict]) -> None:
        for entry in dct:
            pp_name = entry['path']
            for value in entry['values']:
                if 'error_code' in value:
                    val = OperationFailedError(OperationFailedReason(value['error_code']))
                else:
                    val = value['value']
                self.append(pp_name, value['timestamp'], val)

    @Monitor.synchronized
    def to_json(self):
        output = []
        for pathpoint, pathpoint_values in self.items():
            values = []
            for timestamp, value in pathpoint_values:
                if isinstance(value, OperationFailedError):
                    values.append({'timestamp': int(timestamp),
                                   'error_code': value.reason.value})
                else:
                    values.append({'timestamp': int(timestamp),
                                   'value': value})
            output.append({'path': pathpoint,
                           'values': values})
        self.clear()
        self.clear_dirty()
        return output

    def on_read_failed(self, pathpoint: str, error: OperationFailedError,
                       timestamp: tp.Optional[float] = None):
        self.append(pathpoint, timestamp or time.time(), error)




from .base import BaseEventDatabase, BaseEventSynchronization
from .in_memory import InMemoryEventDatabase
from .pickling import PicklingEventDatabase
from .null import NullEventDatabase

__all__ = ['BaseEventDatabase', 'BaseEventSynchronization', 'InMemoryEventDatabase',
           'PicklingEventDatabase', 'NullEventDatabase']



from setuptools import find_packages
from distutils.core import setup
from smok import __version__


setup(version=__version__,
      packages=find_packages(include=['smok', 'smok.*']),
      package_data={'smok': ['certs/dev.crt', 'certs/root.crt']},
      install_requires=[line.strip() for line in open('requirements.txt', 'r').readlines() if line.strip()],
      python_requires='!=2.7.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
      zip_safe=True
      )



from .orders import AdviseLevel, MessageOrder, WaitOrder, WriteOrder, ReadOrder, Section, \
    SysctlOrder
from .pathpoint import Pathpoint
from .typing import PathpointType, PathpointValueType, to_type, ValueOrExcept
from .reparse import ReparsePathpoint

__all__ = ['Pathpoint', 'PathpointValueType', 'AdviseLevel', 'PathpointType',
           'to_type', 'ValueOrExcept', 'MessageOrder', 'WaitOrder', 'ReadOrder',
           'WriteOrder', 'Section', 'ReparsePathpoint', 'SysctlOrder']



import logging
import time

from satella.coding import silence_excs, log_exceptions
from satella.coding.concurrent import PeekableQueue, IntervalTerminableThread
from satella.coding.decorators import retry
from satella.time import time_as_int

from smok.exceptions import ResponseError
from smok.pathpoint.orders import Section
from smok.threads.archives_and_macros.archive import archiving_entries_from_json, \
    ArchivingEntry, archiving_dict_from_json
from smok.macro import Macro

ARCHIVE_UPDATING_INTERVAL = 600
MACROS_UPDATING_INTERVAL = 30 * 60
logger = logging.getLogger(__name__)


class ArchivingAndMacroThread(IntervalTerminableThread):
    def __init__(self, device: 'SMOKDevice', order_queue: PeekableQueue, dont_do_macros,
                 dont_do_archives):
        super().__init__(60)
        self.dont_do_macros = dont_do_macros
        self.dont_do_archives = dont_do_archives
        self.device = device
        self.order_queue = order_queue
        self.archives_updated_on = 0  # type: int
        self.macros_updated_on = 0  # type: int
        self.macros_to_execute = []  # type: tp.List[Macro]
        self.archiving_entries = set()  # type: tp.Set[ArchivingEntry]

        # Load the archiving data
        for interval, pathpoints in self.device.arch_database.get_archiving_instructions().items():
            for pp in pathpoints:
                self.archiving_entries.add(ArchivingEntry(pp, interval))

    def should_update_archives(self) -> bool:
        return time.time() - self.archives_updated_on > ARCHIVE_UPDATING_INTERVAL

    def should_update_macros(self) -> bool:
        return time.time() - self.macros_updated_on > MACROS_UPDATING_INTERVAL

    @retry(3, exc_classes=ResponseError)
    @log_exceptions(logger, logging.ERROR, exc_types=ResponseError)
    def update_macros(self) -> None:
        start = int(self.macros_updated_on)
        if start == 0:
            start = time_as_int() - 2 * MACROS_UPDATING_INTERVAL
        stop = start + 5 * MACROS_UPDATING_INTERVAL
        resp = self.device.api.get('/v1/device/macro/occurrences/%s-%s' % (
            start, stop
        ))
        macros = [Macro.from_json(macro) for macro in resp]
        macros_to_execute = [macro for macro in macros if macro]
        self.device.macros_database.set_macros(macros_to_execute)
        self.macros_updated_on = time.time()

    @retry(3, exc_classes=ResponseError)
    def update_archives(self):
        data = self.device.api.get('/v1/device/pathpoints/archived')
        entries_now = archiving_entries_from_json(data)
        dct = archiving_dict_from_json(data)
        self.device.arch_database.on_archiving_data_sync(dct)
        entries_to_evict = self.archiving_entries - entries_now
        entries_to_add = entries_now - self.archiving_entries
        for entry in entries_to_evict:
            self.archiving_entries.remove(entry)
        for entry in entries_to_add:
            self.archiving_entries.add(entry)
        self.archives_updated_on = time.time()

    def loop(self) -> None:
        if not self.dont_do_macros:
            if self.should_update_macros():
                self.update_macros()

            mdb = self.device.macros_database

            for macro in mdb.get_macros():
                if macro.should_execute():
                    macro.execute(self.device)

            for macro_id, ts in mdb.get_done_macros():
                logger.warning(f'Syncing {macro_id} {ts}')
                with silence_excs(ResponseError):
                    self.device.api.post('/v1/device/macros/%s/%s' % (macro_id, ts))
                    mdb.notify_macro_synced(macro_id, ts)

        self.device.metadata.try_update()

        if not self.dont_do_archives:
            if self.should_update_archives():
                self.update_archives()

            sec = Section()
            for a_entry in self.archiving_entries:
                if a_entry.should_update():
                    sec += a_entry.update()
            if sec:
                self.order_queue.put(sec)
